# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the repair tools service implementation for v0."""

from google.protobuf.wrappers_pb2 import Int32Value
import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.conversions import from_measurement_to_server_angle, from_measurement_to_server_length
from ..base.repair_tools import GRPCRepairToolsService
from .conversions import (
    build_grpc_id,
    serialize_tracker_command_response,
)


class GRPCRepairToolsServiceV0(GRPCRepairToolsService):  # noqa: D102
    """Repair tools service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    Repair Tools service. It is specifically designed for the v0 version
    of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):
        from ansys.api.geometry.v0.repairtools_pb2_grpc import RepairToolsStub

        self.stub = RepairToolsStub(channel)

    @protect_grpc
    def find_split_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindSplitEdgesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindSplitEdgesRequest(
            bodies_or_faces=kwargs["bodies_or_faces"],
            angle=(
                DoubleValue(value=float(from_measurement_to_server_angle(kwargs["angle"])))
                if kwargs["angle"] is not None
                else None
            ),
            distance=(
                DoubleValue(value=float(from_measurement_to_server_length(kwargs["distance"])))
                if kwargs["distance"] is not None
                else None
            ),
        )

        # Call the gRPC service
        response = self.stub.FindSplitEdges(request)

        # Format and return the response as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_extra_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindExtraEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindExtraEdgesRequest(selection=kwargs["selection"])

        # Return the response - formatted as a dictionary
        response = self.stub.FindExtraEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_inexact_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindInexactEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindInexactEdgesRequest(selection=kwargs["selection"])

        # Call the gRPC service
        response = self.stub.FindInexactEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_short_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindShortEdgesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindShortEdgesRequest(
            selection=kwargs["selection"],
            max_edge_length=DoubleValue(value=from_measurement_to_server_length(kwargs["length"])),
        )

        # Call the gRPC service
        response = self.stub.FindShortEdges(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_duplicate_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindDuplicateFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindDuplicateFacesRequest(faces=kwargs["faces"])

        # Call the gRPC service
        response = self.stub.FindDuplicateFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "faces": res.face_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_missing_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindMissingFacesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.geometry.core.logger import LOG

        from ..base.conversions import (
            from_measurement_to_server_angle,
            from_measurement_to_server_length,
        )

        # Check the backend version to set optional parameters
        if kwargs["backend_version"] < (26, 1, 0) and (
            kwargs["angle"] is not None or kwargs["distance"] is not None
        ):
            # If the backend version is less than 26.1.0, set angle and distance to None
            kwargs["angle"] = None
            kwargs["distance"] = None

            # Log a warning
            LOG.warning(
                "The backend version is less than 26.1.0, so angle and distance parameters will be"
                "ignored. Please update the backend to use these parameters."
            )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindMissingFacesRequest(
            faces=kwargs["faces"],
            angle=DoubleValue(value=from_measurement_to_server_angle(kwargs["angle"]))
            if kwargs["angle"] is not None
            else None,
            distance=DoubleValue(value=from_measurement_to_server_length(kwargs["distance"]))
            if kwargs["distance"] is not None
            else None,
        )

        # Call the gRPC service
        response = self.stub.FindMissingFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "edges": res.edge_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_small_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindSmallFacesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.geometry.core.logger import LOG

        from ..base.conversions import (
            from_measurement_to_server_area,
            from_measurement_to_server_length,
        )

        # Check the backend version to set optional parameters
        if kwargs["backend_version"] < (26, 1, 0) and (
            kwargs["area"] is not None or kwargs["width"] is not None
        ):
            # If the backend version is less than 26.1.0, set area and width to None
            kwargs["area"] = None
            kwargs["width"] = None

            # Log a warning
            LOG.warning(
                "The backend version is less than 26.1.0, so area and width parameters will be"
                "ignored. Please update the backend to use these parameters."
            )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindSmallFacesRequest(
            selection=kwargs["selection"],
            area=DoubleValue(value=from_measurement_to_server_area(kwargs["area"]))
            if kwargs["area"] is not None
            else None,
            width=DoubleValue(value=from_measurement_to_server_length(kwargs["width"]))
            if kwargs["width"] is not None
            else None,
        )

        # Call the gRPC service
        response = self.stub.FindSmallFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "faces": res.face_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_stitch_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindStitchFacesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        from ansys.geometry.core.logger import LOG

        from ..base.conversions import from_measurement_to_server_length

        if kwargs["backend_version"] < (26, 1, 0) and kwargs["distance"] is not None:
            # If the backend version is less than 26.1.0, set distance to None and log warning
            kwargs["distance"] = None
            LOG.warning(
                "The backend version is less than 26.1.0, so distance parameter will be ignored. "
            )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindStitchFacesRequest(
            faces=kwargs["faces"],
            maximum_distance=DoubleValue(
                value=from_measurement_to_server_length(kwargs["distance"])
            )
            if kwargs["distance"] is not None
            else None,
        )

        # Call the gRPC service
        response = self.stub.FindStitchFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_simplify(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindAdjustSimplifyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindAdjustSimplifyRequest(selection=kwargs["selection"])

        # Call the gRPC service
        response = self.stub.FindAdjustSimplify(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_interferences(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindInterferenceRequest
        from google.protobuf.wrappers_pb2 import BoolValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindInterferenceRequest(
            bodies=kwargs["bodies"],
            cut_smaller_body=BoolValue(value=kwargs["cut_smaller_body"]),
        )

        # Call the gRPC service
        response = self.stub.FindInterference(request)

        # Return the response - formatted as a dictionary
        return {
            "problems": [
                {
                    "id": res.id,
                    "bodies": res.body_monikers,
                }
                for res in response.result
            ]
        }

    @protect_grpc
    def find_and_fix_short_edges(self, **kwargs):  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindShortEdgesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindShortEdgesRequest(
            selection=kwargs["selection"],
            max_edge_length=DoubleValue(value=from_measurement_to_server_length(kwargs["length"])),
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndFixShortEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def find_and_fix_extra_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindExtraEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindExtraEdgesRequest(
            selection=kwargs["selection"],
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndFixExtraEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": response.created_bodies_monikers,
            "modified_bodies_monikers": response.modified_bodies_monikers,
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def find_and_fix_split_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindSplitEdgesRequest
        from google.protobuf.wrappers_pb2 import DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindSplitEdgesRequest(
            bodies_or_faces=kwargs["bodies_or_faces"],
            distance=DoubleValue(value=float(from_measurement_to_server_length(kwargs["length"]))),
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndFixSplitEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def find_and_fix_simplify(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindAdjustSimplifyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindAdjustSimplifyRequest(
            selection=kwargs["selection"],
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndSimplify(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": [],
            "modified_bodies_monikers": [],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def find_and_fix_stitch_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FindStitchFacesRequest
        from google.protobuf.wrappers_pb2 import BoolValue, DoubleValue

        from ..base.conversions import from_measurement_to_server_length

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindStitchFacesRequest(
            faces=kwargs["body_ids"],
            maximum_distance=DoubleValue(
                value=from_measurement_to_server_length(kwargs["max_distance"])
            )
            if kwargs["max_distance"] is not None
            else None,
            allow_multiple_bodies=BoolValue(value=kwargs["allow_multiple_bodies"]),
            maintain_components=BoolValue(value=kwargs["maintain_components"]),
            check_for_coincidence=BoolValue(value=kwargs["check_for_coincidence"]),
            comprehensive=kwargs["comprehensive_result"],
        )

        # Call the gRPC service
        response = self.stub.FindAndFixStitchFaces(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies_monikers": response.created_bodies_monikers,
            "modified_bodies_monikers": response.modified_bodies_monikers,
            "found": response.found,
            "repaired": response.repaired,
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def inspect_geometry(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import InspectGeometryRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = InspectGeometryRequest(
            bodies=[build_grpc_id(id) for id in kwargs.get("body_ids", [])]
        )

        # Call the gRPC service
        inspect_result_response = self.stub.InspectGeometry(request)

        # Serialize and return the response
        return self.__serialize_inspect_result_response(inspect_result_response)

    @protect_grpc
    def repair_geometry(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import RepairGeometryRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RepairGeometryRequest(
            bodies=[build_grpc_id(id) for id in kwargs.get("body_ids", [])]
        )

        # Call the gRPC service
        response = self.stub.RepairGeometry(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.result.success,
        }

    @protect_grpc
    def fix_duplicate_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixDuplicateFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixDuplicateFacesRequest(
            duplicate_face_problem_area_id=Int32Value(
                value=int(kwargs["duplicate_face_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixDuplicateFaces(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_missing_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixMissingFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixMissingFacesRequest(
            missing_face_problem_area_id=Int32Value(
                value=int(kwargs["missing_face_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixMissingFaces(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_inexact_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixInexactEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixInexactEdgesRequest(
            inexact_edge_problem_area_id=Int32Value(
                value=int(kwargs["inexact_edge_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixInexactEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_extra_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixExtraEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixExtraEdgesRequest(
            extra_edge_problem_area_id=Int32Value(value=int(kwargs["extra_edge_problem_area_id"])),
        )

        # Call the gRPC service
        response = self.stub.FixExtraEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_short_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixShortEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixShortEdgesRequest(
            short_edge_problem_area_id=Int32Value(value=int(kwargs["short_edge_problem_area_id"])),
        )

        # Call the gRPC service
        response = self.stub.FixShortEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_small_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixSmallFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixSmallFacesRequest(
            small_face_problem_area_id=Int32Value(value=int(kwargs["small_face_problem_area_id"])),
        )

        # Call the gRPC service
        response = self.stub.FixSmallFaces(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_split_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixSplitEdgesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixSplitEdgesRequest(
            split_edge_problem_area_id=Int32Value(value=int(kwargs["split_edge_problem_area_id"])),
        )

        # Call the gRPC service
        response = self.stub.FixSplitEdges(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_stitch_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixStitchFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixStitchFacesRequest(
            stitch_face_problem_area_id=Int32Value(
                value=int(kwargs["stitch_face_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixStitchFaces(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_unsimplified_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixAdjustSimplifyRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixAdjustSimplifyRequest(
            adjust_simplify_problem_area_id=Int32Value(
                value=int(kwargs["adjust_simplify_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixAdjustSimplify(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    @protect_grpc
    def fix_interference(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.repairtools_pb2 import FixInterferenceRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = FixInterferenceRequest(
            interference_problem_area_id=Int32Value(
                value=int(kwargs["interference_problem_area_id"])
            ),
        )

        # Call the gRPC service
        response = self.stub.FixInterference(request)

        serialized_tracker_response = serialize_tracker_command_response(
            response=response.result.complete_command_response
        )

        # Return the response - formatted as a dictionary
        return {
            "tracker_response": serialized_tracker_response,
            "success": response.result.success,
            "created_bodies_monikers": response.result.created_bodies_monikers,
            "modified_bodies_monikers": response.result.modified_bodies_monikers,
        }

    def __serialize_inspect_result_response(self, response) -> dict:  # noqa: D102
        def serialize_body(body):
            return {
                "id": body.id,
                "name": body.name,
                "can_suppress": body.can_suppress,
                "transform_to_master": {
                    "m00": body.transform_to_master.m00,
                    "m11": body.transform_to_master.m11,
                    "m22": body.transform_to_master.m22,
                    "m33": body.transform_to_master.m33,
                },
                "master_id": body.master_id,
                "parent_id": body.parent_id,
            }

        def serialize_face(face):
            return {
                "id": face.id,
                "surface_type": face.surface_type,
                "export_id": face.export_id,
                "is_reversed": getattr(face, "is_reversed", False),
                "parent_id": face.parent_id.id,
            }

        def serialize_edge(edge):
            return {
                "id": edge.id,
                "curve_type": edge.curve_type,
                "export_id": edge.export_id,
                "length": edge.length,
                "owner_id": edge.owner_id,
                "parent": serialize_body(edge.parent) if hasattr(edge, "parent") else None,
            }

        def serialize_issue(issue):
            return {
                "message_type": issue.message_type,
                "message_id": issue.message_id,
                "message": issue.message,
                "faces": [serialize_face(f) for f in issue.faces],
                "edges": [serialize_edge(e) for e in issue.edges],
            }

        return {
            "issues_by_body": [
                {
                    "body": serialize_body(body_issues.body),
                    "issues": [serialize_issue(i) for i in body_issues.issues],
                }
                for body_issues in response.issues_by_body
            ]
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""The problem area definition."""

from abc import abstractmethod
from typing import TYPE_CHECKING

import ansys.geometry.core as pyansys_geom
from ansys.geometry.core.connection import GrpcClient
from ansys.geometry.core.misc.auxiliary import (
    get_design_from_body,
    get_design_from_edge,
    get_design_from_face,
)
from ansys.geometry.core.misc.checks import check_type_all_elements_in_iterable
from ansys.geometry.core.tools.repair_tool_message import (
    RepairToolMessage,
    create_repair_message_from_response,
)

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.body import Body
    from ansys.geometry.core.designer.edge import Edge
    from ansys.geometry.core.designer.face import Face


class ProblemArea:
    """Represents problem areas.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    """

    def __init__(self, id: str, grpc_client: GrpcClient):
        """Initialize a new instance of a problem area class."""
        self._id = id
        self._grpc_client = grpc_client

    @property
    def id(self) -> str:
        """The id of the problem area."""
        return self._id

    @abstractmethod
    def fix(self):
        """Fix problem area."""
        raise NotImplementedError("Fix method is not implemented in the base class.")

    def build_repair_tool_message(self, response: dict) -> RepairToolMessage:
        """Build a repair tool message from the service response.

        Parameters
        ----------
        response : dict
            The response from the service containing information about the repair operation.

        Returns
        -------
        RepairToolMessage
            A message containing the success status, created bodies, modified bodies,
            number of found problem areas, and number of repaired problem areas.
        """
        return RepairToolMessage(
            success=response.get("success"),
            created_bodies=response.get("created_bodies_monikers", []),
            modified_bodies=response.get("modified_bodies_monikers", []),
            found=response.get("found", -1),
            repaired=response.get("repaired", -1),
        )


class DuplicateFaceProblemAreas(ProblemArea):
    """Provides duplicate face problem area definition.

    Represents a duplicate face problem area with unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[Face]
        List of faces associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, faces: list["Face"]):
        """Initialize a new instance of the duplicate face problem areaclass."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.face import Face

        # Verify that all elements in the list are of type Face
        check_type_all_elements_in_iterable(faces, Face)

        self._faces = faces

    @property
    def faces(self) -> list["Face"]:
        """The list of faces connected to this problem area."""
        return self._faces

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.faces:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_face(self.faces[0])
        response = self._grpc_client.services.repair_tools.fix_duplicate_faces(
            duplicate_face_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class MissingFaceProblemAreas(ProblemArea):
    """Provides missing face problem area definition.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the missing face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._grpc_client.services.repair_tools.fix_missing_faces(
            missing_face_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class InexactEdgeProblemAreas(ProblemArea):
    """Represents an inexact edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the inexact edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])

        response = self._grpc_client.services.repair_tools.fix_inexact_edges(
            inexact_edge_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class ExtraEdgeProblemAreas(ProblemArea):
    """Represents a extra edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the extra edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._grpc_client.services.repair_tools.fix_extra_edges(
            extra_edge_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class ShortEdgeProblemAreas(ProblemArea):
    """Represents a short edge problem area with a unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the ``ShortEdgeProblemAreas`` class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are edges
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._grpc_client.services.repair_tools.fix_short_edges(
            short_edge_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class SmallFaceProblemAreas(ProblemArea):
    """Represents a small face problem area with a unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[Face]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, faces: list["Face"]):
        """Initialize a new instance of the small face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.face import Face

        # Verify that all elements in the list are of type Face
        check_type_all_elements_in_iterable(faces, Face)

        self._faces = faces

    @property
    def faces(self) -> list["Face"]:
        """The list of faces connected to this problem area."""
        return self._faces

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.faces:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_face(self.faces[0])
        response = self._grpc_client.services.repair_tools.fix_small_faces(
            small_face_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class SplitEdgeProblemAreas(ProblemArea):
    """Represents a split edge problem area with unique identifier and associated edges.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    edges : list[Edge]
        List of edges associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, edges: list["Edge"]):
        """Initialize a new instance of the split edge problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.edge import Edge

        # Verify that all elements in the list are of type Edge
        check_type_all_elements_in_iterable(edges, Edge)

        self._edges = edges

    @property
    def edges(self) -> list["Edge"]:
        """The list of edges connected to this problem area."""
        return self._edges

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.edges:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_edge(self.edges[0])
        response = self._grpc_client.services.repair_tools.fix_split_edges(
            split_edge_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class StitchFaceProblemAreas(ProblemArea):
    """Represents a stitch face problem area with unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    bodies : list[Body]
        List of bodies associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, bodies: list["Body"]):
        """Initialize a new instance of the stitch face problem area class."""
        super().__init__(id, grpc_client)

        from ansys.geometry.core.designer.body import Body

        # Verify that all elements in the list are of type Body
        check_type_all_elements_in_iterable(bodies, Body)

        self._bodies = bodies

    @property
    def bodies(self) -> list["Body"]:
        """The list of bodies connected to this problem area."""
        return self._bodies

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.bodies:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_body(self.bodies[0])
        response = self._grpc_client.services.repair_tools.fix_stitch_faces(
            stitch_face_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class UnsimplifiedFaceProblemAreas(ProblemArea):
    """Represents a unsimplified face problem area with unique identifier and associated faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[Face]
        List of faces associated with the design.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, faces: list["Face"]):
        """Initialize a new instance of the unsimplified face problem area class."""
        super().__init__(id, grpc_client)

        self._faces = faces

    @property
    def faces(self) -> list["Face"]:
        """The list of faces connected to this problem area."""
        return self._faces

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.
        """
        if not self.faces:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_face(self.faces[0])
        response = self._grpc_client.services.repair_tools.fix_unsimplified_faces(
            adjust_simplify_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        message = create_repair_message_from_response(response)
        return message


class InterferenceProblemAreas(ProblemArea):
    """Represents an interference problem area with a unique identifier and associated bodies.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    bodies : list[Body]
        List of bodies in the problem area.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, bodies: list["Body"]):
        """Initialize a new instance of the interference problem area class."""
        super().__init__(id, grpc_client)

        self._bodies = bodies

    @property
    def bodies(self) -> list["Body"]:
        """The list of the bodies connected to this problem area."""
        return self._bodies

    def fix(self) -> RepairToolMessage:
        """Fix the problem area.

        Returns
        -------
        message: RepairToolMessage
            Message containing created and/or modified bodies.

        Notes
        -----
        The current implementation does not properly track changes.
        The list of created and modified bodies are empty.
        """
        if not self.bodies:
            return RepairToolMessage(False, [], [])

        parent_design = get_design_from_body(self.bodies[0])
        response = self._grpc_client.services.repair_tools.fix_interference(
            interference_problem_area_id=self.id
        )

        if not pyansys_geom.USE_TRACKER_TO_UPDATE_DESIGN:
            parent_design._update_design_inplace()
        else:
            parent_design._update_from_tracker(response.get("tracker_response"))

        ## The tool does not return the created or modified objects.
        ## https://github.com/ansys/pyansys-geometry/issues/1319
        message = RepairToolMessage(response.get("success"), [], [])
        return message


class LogoProblemArea(ProblemArea):
    """Represents a logo problem area defined by a list of faces.

    Parameters
    ----------
    id : str
        Server-defined ID for the problem area.
    grpc_client : GrpcClient
        Active supporting geometry service instance for design modeling.
    faces : list[str]
        List of faces defining the logo problem area.
    """

    def __init__(self, id: str, grpc_client: GrpcClient, face_ids: list[str]):
        """Initialize a new instance of the logo problem area class."""
        super().__init__(id, grpc_client)

        self._face_ids = face_ids

    @property
    def face_ids(self) -> list[str]:
        """The ids of the faces defining the logos."""
        return self._face_ids

    def fix(self) -> bool:
        """Fix the problem area by deleting the logos.

        Returns
        -------
        message: bool
            Message that return whether the operation was successful.
        """
        if len(self._face_ids) == 0:
            return False

        response = self._grpc_client._services.prepare_tools.remove_logo(face_ids=self._face_ids)

        return response.get("success")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""PyAnsys Geometry math subpackage."""

from ansys.geometry.core.math.bbox import BoundingBox
from ansys.geometry.core.math.constants import (
    DEFAULT_POINT2D,
    DEFAULT_POINT3D,
    IDENTITY_MATRIX33,
    IDENTITY_MATRIX44,
    UNITVECTOR2D_X,
    UNITVECTOR2D_Y,
    UNITVECTOR3D_X,
    UNITVECTOR3D_Y,
    UNITVECTOR3D_Z,
    ZERO_POINT2D,
    ZERO_POINT3D,
    ZERO_VECTOR2D,
    ZERO_VECTOR3D,
)
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.matrix import Matrix, Matrix33, Matrix44
from ansys.geometry.core.math.misc import get_two_circle_intersections
from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point2D, Point3D
from ansys.geometry.core.math.vector import UnitVector2D, UnitVector3D, Vector2D, Vector3D



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module to perform a connection validation check.

The method in this module is only used for testing the default Docker service on
GitHub and can safely be skipped within testing.

This command shows how this method is typically used:

.. code:: bash

   python -c "from ansys.geometry.core.connection import validate; validate()"
"""

from ansys.geometry.core.connection.client import wait_until_healthy
import ansys.geometry.core.connection.defaults as default_settings


def validate(*args, **kwargs):  # pragma: no cover
    """Create a client using the default settings and validate it."""
    # Assume local transport mode for validation if not provided
    channel = None
    if "transport_mode" not in kwargs:
        import platform

        try:
            channel = wait_until_healthy(
                channel=f"{kwargs.get('host', default_settings.DEFAULT_HOST)}:{kwargs.get('port', default_settings.DEFAULT_PORT)}",  # noqa: E501
                timeout=kwargs.get("timeout", 120),
                transport_mode="wnua" if platform.system() == "Windows" else "uds",
            )
        except Exception:
            # Let's give it a try to insecure mode... just in case
            channel = wait_until_healthy(
                channel=f"{kwargs.get('host', default_settings.DEFAULT_HOST)}:{kwargs.get('port', default_settings.DEFAULT_PORT)}",  # noqa: E501
                timeout=kwargs.get("timeout", 120),
                transport_mode="insecure",
            )
    else:
        channel = wait_until_healthy(
            channel=f"{kwargs.get('host', default_settings.DEFAULT_HOST)}:{kwargs.get('port', default_settings.DEFAULT_PORT)}",  # noqa: E501
            timeout=kwargs.get("timeout", 120),
            transport_mode=kwargs["transport_mode"],
            uds_dir=kwargs.get("uds_dir", None),
            uds_id=kwargs.get("uds_id", None),
            certs_dir=kwargs.get("certs_dir", None),
        )

    # If we reach this point, the connection is valid
    print("Connection to Geometry server is valid.")
    print(f"Using gRPC channel connected to {channel._channel.target().decode()}")

    # TODO: consider adding additional server stat reporting
    # https://github.com/ansys/pyansys-geometry/issues/1319



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing v1 related conversions from PyAnsys Geometry objects to gRPC messages."""

from typing import TYPE_CHECKING

from ansys.api.discovery.v1.commands.file_pb2 import (
    ImportOptionDefinition as GRPCImportOptionDefinition,
)
from ansys.api.discovery.v1.commonenums_pb2 import (
    BackendType as GRPCBackendType,
    FileFormat as GRPCFileFormat,
)
from ansys.api.discovery.v1.commonmessages_pb2 import (
    Arc as GRPCArc,
    Circle as GRPCCircle,
    Direction as GRPCDirection,
    Ellipse as GRPCEllipse,
    EntityIdentifier,
    Frame as GRPCFrame,
    Line as GRPCLine,
    Plane as GRPCPlane,
    Point as GRPCPoint,
    Polygon as GRPCPolygon,
    Quantity as GRPCQuantity,
)
from ansys.api.discovery.v1.design.designmessages_pb2 import (
    BodyEntity as GRPCBodyEntity,
    ComponentEntity as GRPCComponentEntity,
    CurveGeometry as GRPCCurveGeometry,
    DatumPointEntity as GRPCDesignPoint,
    DrivingDimensionEntity as GRPCDrivingDimension,
    EdgeTessellation as GRPCEdgeTessellation,
    Geometries as GRPCGeometries,
    Knot as GRPCKnot,
    MaterialEntity as GRPCMaterial,
    MaterialProperty as GRPCMaterialProperty,
    Matrix as GRPCMatrix,
    NurbsCurve as GRPCNurbsCurve,
    NurbsSurface as GRPCNurbsSurface,
    PartEntity as GRPCPartEntity,
    Surface as GRPCSurface,
    Tessellation as GRPCTessellation,
    TessellationOptions as GRPCTessellationOptions,
    TrackedChanges as GRPCTrackedChanges,
    TrackedCommandResponse as GRPCTrackedCommandResponse,
    TrimmedCurve as GRPCTrimmedCurve,
    TrimmedSurface as GRPCTrimmedSurface,
)
from ansys.api.discovery.v1.design.parameters.drivingdimension_pb2 import (
    UpdateStatus as GRPCUpdateStatus,
)
from ansys.api.discovery.v1.geometryenums_pb2 import (
    SurfaceType as GRPCSurfaceType,
)
from ansys.api.discovery.v1.operations.prepare_pb2 import (
    EnclosureOptions as GRPCEnclosureOptions,
)
from ansys.api.discovery.v1.operations.repair_pb2 import (
    RepairToolMessage as GRPCRepairToolResponse,
)
import pint

from ansys.geometry.core.errors import GeometryRuntimeError
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS, Distance
from ansys.geometry.core.shapes.surfaces.nurbs import NURBSSurface

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv
    import semver

    from ansys.geometry.core.connection.backend import BackendType
    from ansys.geometry.core.designer.design import DesignFileFormat
    from ansys.geometry.core.designer.face import SurfaceType
    from ansys.geometry.core.materials.material import Material
    from ansys.geometry.core.materials.property import MaterialProperty
    from ansys.geometry.core.math.frame import Frame
    from ansys.geometry.core.math.matrix import Matrix44
    from ansys.geometry.core.math.plane import Plane
    from ansys.geometry.core.math.point import Point2D, Point3D
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.misc.measurements import Measurement
    from ansys.geometry.core.misc.options import ImportOptionsDefinitions, TessellationOptions
    from ansys.geometry.core.parameters.parameter import (
        Parameter,
        ParameterUpdateStatus,
    )
    from ansys.geometry.core.shapes.curves.curve import Curve
    from ansys.geometry.core.shapes.curves.line import Line
    from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve
    from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
    from ansys.geometry.core.shapes.surfaces.surface import Surface
    from ansys.geometry.core.shapes.surfaces.trimmed_surface import TrimmedSurface
    from ansys.geometry.core.sketch.arc import Arc
    from ansys.geometry.core.sketch.circle import SketchCircle
    from ansys.geometry.core.sketch.edge import SketchEdge
    from ansys.geometry.core.sketch.ellipse import SketchEllipse
    from ansys.geometry.core.sketch.face import SketchFace
    from ansys.geometry.core.sketch.nurbs import SketchNurbs
    from ansys.geometry.core.sketch.polygon import Polygon
    from ansys.geometry.core.sketch.segment import SketchSegment
    from ansys.geometry.core.tools.prepare_tools import EnclosureOptions


def from_grpc_backend_type_to_backend_type(
    grpc_backend_type: GRPCBackendType,
) -> "BackendType":
    """Convert a gRPC v1 backend type to a backend type.

    Parameters
    ----------
    backend_type : GRPCBackendType
        Source backend type.

    Returns
    -------
    BackendType
        Converted backend type.
    """
    from ansys.geometry.core.connection.backend import BackendType

    # Map the gRPC backend type to the corresponding BackendType
    backend_type = None

    if grpc_backend_type == GRPCBackendType.BACKENDTYPE_DISCOVERY:
        backend_type = BackendType.DISCOVERY
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_SPACECLAIM:
        backend_type = BackendType.SPACECLAIM
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_WINDOWS_DMS:
        backend_type = BackendType.WINDOWS_SERVICE
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_LINUX_DMS:
        backend_type = BackendType.LINUX_SERVICE
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_CORE_SERVICE_LINUX:
        backend_type = BackendType.CORE_LINUX
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_CORE_SERVICE_WINDOWS:
        backend_type = BackendType.CORE_WINDOWS
    elif grpc_backend_type == GRPCBackendType.BACKENDTYPE_DISCOVERY_HEADLESS:
        backend_type = BackendType.DISCOVERY_HEADLESS
    else:
        raise ValueError(f"Invalid backend type: {grpc_backend_type}")

    return backend_type


def build_grpc_id(id: str) -> EntityIdentifier:
    """Build a v1 EntityIdentifier gRPC message.

    Parameters
    ----------
    id : str
        Source ID.

    Returns
    -------
    EntityIdentifier
        Geometry service gRPC entity identifier message.
    """
    return EntityIdentifier(id=id)


def from_point3d_to_grpc_point(point: "Point3D") -> GRPCPoint:
    """Convert a v1 ``Point3D`` class to a point gRPC message.

    Parameters
    ----------
    point : Point3D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    return GRPCPoint(
        x=GRPCQuantity(value_in_geometry_units=point.x.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
        y=GRPCQuantity(value_in_geometry_units=point.y.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
        z=GRPCQuantity(value_in_geometry_units=point.z.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
    )


def from_grpc_point_to_point3d(point: GRPCPoint) -> "Point3D":
    """Convert a v1 point gRPC message class to a ``Point3D`` class.

    Parameters
    ----------
    point : GRPCPoint
        Source point data.

    Returns
    -------
    Point3D
        Converted point.
    """
    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return Point3D(
        [
            point.x.value_in_geometry_units,
            point.y.value_in_geometry_units,
            point.z.value_in_geometry_units,
        ],
        DEFAULT_UNITS.SERVER_LENGTH,
    )


def from_point2d_to_grpc_point(plane: "Plane", point2d: "Point2D") -> GRPCPoint:
    """Convert a v1 ``Point2D`` class to a point gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D point.
    point : Point2D
        Source point data.

    Returns
    -------
    GRPCPoint
        Geometry service gRPC point message. The unit is meters.
    """
    point3d = plane.transform_point2d_local_to_global(point2d)
    return GRPCPoint(
        x=GRPCQuantity(value_in_geometry_units=point3d.x.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
        y=GRPCQuantity(value_in_geometry_units=point3d.y.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
        z=GRPCQuantity(value_in_geometry_units=point3d.z.m_as(DEFAULT_UNITS.SERVER_LENGTH)),
    )


def from_point3d_to_grpc_design_point(point: "Point3D") -> GRPCDesignPoint:
    """Convert a v1 ``Point3D`` class to a design point gRPC message.

    Parameters
    ----------
    point : Point3D
        Source point data.

    Returns
    -------
    GRPCDesignPoint
        Geometry service gRPC design point message. The unit is meters.
    """
    return GRPCDesignPoint(
        position=from_point3d_to_grpc_point(point),
    )


def from_unit_vector_to_grpc_direction(unit_vector: "UnitVector3D") -> GRPCDirection:
    """Convert a v1 ``UnitVector3D`` class to a unit vector gRPC message.

    Parameters
    ----------
    unit_vector : UnitVector3D
        Source vector data.

    Returns
    -------
    GRPCDirection
        Geometry service gRPC direction message.
    """
    return GRPCDirection(x=unit_vector.x, y=unit_vector.y, z=unit_vector.z)


def from_line_to_grpc_line(line: "Line") -> GRPCLine:
    """Convert a v1 ``Line`` to a line gRPC message.

    Parameters
    ----------
    line : Line
        Line to convert.

    Returns
    -------
    GRPCLine
        Geometry service gRPC ``Line`` message.
    """
    start = line.origin
    end = line.origin + line.direction
    return GRPCLine(start=from_point3d_to_grpc_point(start), end=from_point3d_to_grpc_point(end))


def from_grpc_material_to_material(material: GRPCMaterial) -> "Material":
    """Convert a v1 material gRPC message to a ``Material`` class.

    Parameters
    ----------
    material : GRPCMaterial
        Material gRPC message.

    Returns
    -------
    Material
        Converted material.
    """
    from ansys.geometry.core.materials.material import Material
    from ansys.geometry.core.materials.property import MaterialPropertyType
    from ansys.geometry.core.misc.units import UNITS

    properties = []
    density = pint.Quantity(0, UNITS.kg / UNITS.m**3)
    for property in material.material_properties:
        mp = from_grpc_material_property_to_material_property(property)
        properties.append(mp)
        if mp.type == MaterialPropertyType.DENSITY:
            density = mp.quantity

    return Material(material.name, density, properties)


def from_grpc_material_property_to_material_property(
    material_property: GRPCMaterialProperty,
) -> "MaterialProperty":
    """Convert a v1 material property gRPC message to a ``MaterialProperty`` class.

    Parameters
    ----------
    material_property : GRPCMaterialProperty
        Material property gRPC message.

    Returns
    -------
    MaterialProperty
        Converted material property.
    """
    from ansys.geometry.core.materials.property import MaterialProperty, MaterialPropertyType

    # In v1 API, id is an EntityIdentifier
    property_id = material_property.id.id

    try:
        mp_type = MaterialPropertyType.from_id(property_id)
    except ValueError:
        mp_type = property_id

    try:
        mp_quantity = pint.Quantity(material_property.value, material_property.units)
    except (
        pint.UndefinedUnitError,
        TypeError,
    ):
        mp_quantity = material_property.value

    return MaterialProperty(mp_type, material_property.display_name, mp_quantity)


def from_frame_to_grpc_frame(frame: "Frame") -> GRPCFrame:
    """Convert a v1 ``Frame`` class to a frame gRPC message.

    Parameters
    ----------
    frame : Frame
        Source frame data.

    Returns
    -------
    GRPCFrame
        Geometry service gRPC frame message. The unit for the frame origin is meters.
    """
    return GRPCFrame(
        origin=from_point3d_to_grpc_point(frame.origin),
        dir_x=from_unit_vector_to_grpc_direction(frame.direction_x),
        dir_y=from_unit_vector_to_grpc_direction(frame.direction_y),
    )


def from_grpc_frame_to_frame(frame: GRPCFrame) -> "Frame":
    """Convert a v1 frame gRPC message to a ``Frame`` class.

    Parameters
    ----------
    frame : GRPCFrame
        Source frame data.

    Returns
    -------
    Frame
        Converted frame.
    """
    from ansys.geometry.core.math.frame import Frame
    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return Frame(
        Point3D(
            input=[
                frame.origin.x.value_in_geometry_units,
                frame.origin.y.value_in_geometry_units,
                frame.origin.z.value_in_geometry_units,
            ],
            unit=DEFAULT_UNITS.SERVER_LENGTH,
        ),
        UnitVector3D(
            input=[
                frame.dir_x.x,
                frame.dir_x.y,
                frame.dir_x.z,
            ]
        ),
        UnitVector3D(
            input=[
                frame.dir_y.x,
                frame.dir_y.y,
                frame.dir_y.z,
            ]
        ),
    )


def from_plane_to_grpc_plane(plane: "Plane") -> GRPCPlane:
    """Convert a v1 ``Plane`` class to a plane gRPC message.

    Parameters
    ----------
    plane : Plane
        Source plane data.

    Returns
    -------
    GRPCPlane
        Geometry service gRPC plane message. The unit is meters.
    """
    return GRPCPlane(
        frame=GRPCFrame(
            origin=from_point3d_to_grpc_point(plane.origin),
            dir_x=from_unit_vector_to_grpc_direction(plane.direction_x),
            dir_y=from_unit_vector_to_grpc_direction(plane.direction_y),
        )
    )


def from_grpc_plane_to_plane(plane: GRPCPlane) -> "Plane":
    """Convert a v1 plane gRPC message to a ``Plane`` class.

    Parameters
    ----------
    plane : GRPCPlane
        Source plane data.

    Returns
    -------
    Plane
        Converted plane.
    """
    from ansys.geometry.core.math.plane import Plane

    frame = from_grpc_frame_to_frame(plane.frame)
    return Plane(frame.origin, frame.direction_x, frame.direction_y)


@graphics_required
def from_grpc_tess_to_pd(tess: GRPCTessellation) -> "pv.PolyData":
    """Convert a v1 ``Tessellation`` to ``pyvista.PolyData``."""
    # lazy imports here to improve initial load
    import numpy as np
    import pyvista as pv

    if len(tess.faces) == 0 or len(tess.vertices) == 0:
        return pv.PolyData()

    return pv.PolyData(var_inp=np.array(tess.vertices).reshape(-1, 3), faces=tess.faces)


def from_grpc_tess_to_raw_data(tess: GRPCTessellation) -> dict:
    """Convert a v1 ``Tessellation`` to raw data."""
    return {"vertices": tess.vertices, "faces": tess.faces, "is_edge": False}


@graphics_required
def from_grpc_edge_tess_to_pd(tess: GRPCEdgeTessellation) -> "pv.PolyData":
    """Convert a v1 ``EdgeTessellation`` to ``pyvista.PolyData``."""
    # lazy imports here to improve initial load
    import numpy as np
    import pyvista as pv

    if len(tess.vertices) == 0:
        return pv.PolyData()

    points = np.reshape(np.array([from_grpc_point_to_point3d(pt) for pt in tess.vertices]), (-1, 3))
    lines = np.hstack([[len(points), *range(len(points))]])
    return pv.PolyData(points, lines=lines)


def from_grpc_edge_tess_to_raw_data(tess: GRPCEdgeTessellation) -> dict:
    """Convert a v1 ``EdgeTessellation`` to raw data."""
    return {
        "vertices": [
            coord.value_in_geometry_units for pt in tess.vertices for coord in (pt.x, pt.y, pt.z)
        ],
        "is_edge": True,
    }


def from_tess_options_to_grpc_tess_options(
    options: "TessellationOptions",
) -> GRPCTessellationOptions:
    """Convert a v1 ``TessellationOptions`` class to a tessellation options gRPC message.

    Parameters
    ----------
    options : TessellationOptions
        Source tessellation options data.

    Returns
    -------
    GRPCTessellationOptions
        Geometry service gRPC tessellation options message.
    """
    from ansys.geometry.core.misc.measurements import DEFAULT_UNITS

    return GRPCTessellationOptions(
        surface_deviation=GRPCQuantity(
            value_in_geometry_units=options.surface_deviation.value.m_as(
                DEFAULT_UNITS.SERVER_LENGTH
            )
        ),
        angle_deviation=GRPCQuantity(
            value_in_geometry_units=options.angle_deviation.value.m_as(DEFAULT_UNITS.SERVER_ANGLE)
        ),
        curve_deviation=GRPCQuantity(value_in_geometry_units=0),
        maximum_aspect_ratio=GRPCQuantity(value_in_geometry_units=options.max_aspect_ratio),
        maximum_edge_length=GRPCQuantity(
            value_in_geometry_units=options.max_edge_length.value.m_as(DEFAULT_UNITS.SERVER_LENGTH)
        ),
        watertight=options.watertight,
    )


def from_sketch_shapes_to_grpc_geometries(
    plane: "Plane",
    edges: list["SketchEdge"],
    faces: list["SketchFace"],
    only_one_curve: bool = False,
) -> GRPCGeometries:
    """Convert lists of ``SketchEdge`` and ``SketchFace`` to a gRPC message.

    Parameters
    ----------
    plane : Plane
        Plane for positioning the 2D sketches.
    edges : list[SketchEdge]
        Source edge data.
    faces : list[SketchFace]
        Source face data.
    only_one_curve : bool, default: False
        Whether to project one curve of the whole set of geometries to
        enhance performance.

    Returns
    -------
    GRPCGeometries
        Geometry service gRPC geometries message. The unit is meters.
    """
    from ansys.geometry.core.sketch.circle import SketchCircle
    from ansys.geometry.core.sketch.ellipse import SketchEllipse
    from ansys.geometry.core.sketch.polygon import Polygon

    geometries = GRPCGeometries()

    converted_sketch_edges = from_sketch_edges_to_grpc_geometries(edges, plane)
    geometries.lines.extend(converted_sketch_edges[0])
    geometries.arcs.extend(converted_sketch_edges[1])
    geometries.nurbs_curves.extend(converted_sketch_edges[2])

    for face in faces:
        if isinstance(face, SketchCircle):
            geometries.circles.append(from_sketch_circle_to_grpc_circle(face, plane))
        if isinstance(face, SketchEllipse):
            geometries.ellipses.append(from_sketch_ellipse_to_grpc_ellipse(face, plane))
        if isinstance(face, Polygon):
            geometries.polygons.append(from_sketch_polygon_to_grpc_polygon(face, plane))
        else:
            converted_face_edges = from_sketch_edges_to_grpc_geometries(face.edges, plane)
            geometries.lines.extend(converted_face_edges[0])
            geometries.arcs.extend(converted_face_edges[1])

    if only_one_curve:
        one_curve_geometry = GRPCGeometries()
        if len(geometries.lines) > 0:
            one_curve_geometry.lines.append(geometries.lines[0])
        elif len(geometries.arcs) > 0:
            one_curve_geometry.arcs.append(geometries.arcs[0])
        elif len(geometries.circles) > 0:
            one_curve_geometry.circles.append(geometries.circles[0])
        elif len(geometries.ellipses) > 0:
            one_curve_geometry.ellipses.append(geometries.ellipses[0])
        elif len(geometries.polygons) > 0:
            one_curve_geometry.polygons.append(geometries.polygons[0])
        elif len(geometries.nurbs_curves) > 0:
            one_curve_geometry.nurbs_curves.append(geometries.nurbs_curves[0])
        return one_curve_geometry

    else:
        return geometries


def from_sketch_edges_to_grpc_geometries(
    edges: list["SketchEdge"],
    plane: "Plane",
) -> tuple[list[GRPCLine], list[GRPCArc], list[GRPCNurbsCurve]]:
    """Convert a v1 list of ``SketchEdge`` to a gRPC message.

    Parameters
    ----------
    edges : list[SketchEdge]
        Source edge data.
    plane : Plane
        Plane for positioning the 2D sketches.

    Returns
    -------
    tuple[list[GRPCLine], list[GRPCArc], list[GRPCNurbsCurve]]
        Geometry service gRPC line, arc, and NURBS curve messages. The unit is meters.
    """
    from ansys.geometry.core.sketch.arc import Arc
    from ansys.geometry.core.sketch.nurbs import SketchNurbs
    from ansys.geometry.core.sketch.segment import SketchSegment

    arcs = []
    segments = []
    nurbs_curves = []
    for edge in edges:
        if isinstance(edge, SketchSegment):
            segments.append(from_sketch_segment_to_grpc_line(edge, plane))
        elif isinstance(edge, Arc):
            arcs.append(from_sketch_arc_to_grpc_arc(edge, plane))
        elif isinstance(edge, SketchNurbs):
            nurbs_curves.append(from_sketch_nurbs_to_grpc_nurbs_curve(edge, plane))

    return (segments, arcs, nurbs_curves)


def from_sketch_arc_to_grpc_arc(arc: "Arc", plane: "Plane") -> GRPCArc:
    """Convert a v1 ``Arc`` class to an arc gRPC message.

    Parameters
    ----------
    arc : Arc
        Source arc data.
    plane : Plane
        Plane for positioning the arc within.

    Returns
    -------
    GRPCArc
        Geometry service gRPC arc message. The unit is meters.
    """
    axis = (
        from_unit_vector_to_grpc_direction(plane.direction_z)
        if not arc.is_clockwise
        else from_unit_vector_to_grpc_direction(-plane.direction_z)
    )

    return GRPCArc(
        center=from_point2d_to_grpc_point(plane, arc.center),
        start=from_point2d_to_grpc_point(plane, arc.start),
        end=from_point2d_to_grpc_point(plane, arc.end),
        axis=axis,
    )


def from_sketch_nurbs_to_grpc_nurbs_curve(curve: "SketchNurbs", plane: "Plane") -> GRPCNurbsCurve:
    """Convert a v1 ``SketchNurbs`` class to a NURBS curve gRPC message.

    Parameters
    ----------
    nurbs : SketchNurbs
        Source NURBS data.
    plane : Plane
        Plane for positioning the NURBS curve.

    Returns
    -------
    GRPCNurbsCurve
        Geometry service gRPC NURBS curve message. The unit is meters.
    """
    from ansys.api.discovery.v1.design.designmessages_pb2 import (
        ControlPoint as GRPCControlPoint,
        NurbsData as GRPCNurbsData,
    )

    # Convert control points
    control_points = [
        GRPCControlPoint(
            position=from_point2d_to_grpc_point(plane, pt),
            weight=curve.weights[i],
        )
        for i, pt in enumerate(curve.control_points)
    ]

    # Convert nurbs data
    nurbs_data = GRPCNurbsData(
        degree=curve.degree,
        knots=from_knots_to_grpc_knots(curve.knots),
        order=curve.degree + 1,
    )

    return GRPCNurbsCurve(
        control_points=control_points,
        nurbs_data=nurbs_data,
    )


def from_sketch_ellipse_to_grpc_ellipse(ellipse: "SketchEllipse", plane: "Plane") -> GRPCEllipse:
    """Convert a v1 ``SketchEllipse`` class to an ellipse gRPC message.

    Parameters
    ----------
    ellipse : SketchEllipse
        Source ellipse data.

    Returns
    -------
    GRPCEllipse
        Geometry service gRPC ellipse message. The unit is meters.
    """
    return GRPCEllipse(
        center=from_point2d_to_grpc_point(plane, ellipse.center),
        majorradius=GRPCQuantity(
            value_in_geometry_units=ellipse.major_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
        ),  # noqa: E501
        minorradius=GRPCQuantity(
            value_in_geometry_units=ellipse.minor_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
        ),  # noqa: E501
        angle=GRPCQuantity(value_in_geometry_units=ellipse.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE)),
    )


def from_sketch_circle_to_grpc_circle(circle: "SketchCircle", plane: "Plane") -> GRPCCircle:
    """Convert a v1 ``SketchCircle`` class to a circle gRPC message.

    Parameters
    ----------
    circle : SketchCircle
        Source circle data.
    plane : Plane
        Plane for positioning the circle.

    Returns
    -------
    GRPCCircle
        Geometry service gRPC circle message. The unit is meters.
    """
    return GRPCCircle(
        center=from_point2d_to_grpc_point(plane, circle.center),
        radius=GRPCQuantity(
            value_in_geometry_units=circle.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
        ),
    )


def from_sketch_polygon_to_grpc_polygon(polygon: "Polygon", plane: "Plane") -> GRPCPolygon:
    """Convert a v1 ``Polygon`` class to a polygon gRPC message.

    Parameters
    ----------
    polygon : Polygon
        Source polygon data.

    Returns
    -------
    GRPCPolygon
        Geometry service gRPC polygon message. The unit is meters.
    """
    return GRPCPolygon(
        center=from_point2d_to_grpc_point(plane, polygon.center),
        radius=GRPCQuantity(
            value_in_geometry_units=polygon.inner_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
        ),  # noqa: E501
        numberofsides=polygon.n_sides,
        angle=GRPCQuantity(value_in_geometry_units=polygon.angle.m_as(DEFAULT_UNITS.SERVER_ANGLE)),
    )


def from_sketch_segment_to_grpc_line(segment: "SketchSegment", plane: "Plane") -> GRPCLine:
    """Convert a v1 ``Segment`` class to a line gRPC message.

    Parameters
    ----------
    segment : SketchSegment
        Source segment data.

    Returns
    -------
    GRPCLine
        Geometry service gRPC line message. The unit is meters.
    """
    return GRPCLine(
        start=from_point2d_to_grpc_point(plane, segment.start),
        end=from_point2d_to_grpc_point(plane, segment.end),
    )


def from_trimmed_curve_to_grpc_trimmed_curve(curve: "TrimmedCurve") -> GRPCTrimmedCurve:
    """Convert a v1 ``TrimmedCurve`` to a trimmed curve gRPC message.

    Parameters
    ----------
    curve : TrimmedCurve
        Curve to convert.

    Returns
    -------
    GRPCTrimmedCurve
        Geometry service gRPC ``TrimmedCurve`` message.
    """
    curve_geometry = from_curve_to_grpc_curve(curve.geometry)
    i_start = curve.interval.start
    i_end = curve.interval.end

    return GRPCTrimmedCurve(
        curve=curve_geometry,
        interval_start=i_start,
        interval_end=i_end,
    )


def from_curve_to_grpc_curve(curve: "Curve") -> GRPCCurveGeometry:
    """Convert a v1 ``Curve`` object to a curve gRPC message.

    Parameters
    ----------
    curve : Curve
        Curve to convert.

    Returns
    -------
    GRPCCurve
        Return ``Curve`` as a ``ansys.api.geometry.CurveGeometry`` message.
    """
    from ansys.geometry.core.shapes.curves.circle import Circle
    from ansys.geometry.core.shapes.curves.ellipse import Ellipse
    from ansys.geometry.core.shapes.curves.line import Line
    from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve

    grpc_curve = None

    if isinstance(curve, Line):
        origin = from_point3d_to_grpc_point(curve.origin)
        direction = from_unit_vector_to_grpc_direction(curve.direction)
        grpc_curve = GRPCCurveGeometry(origin=origin, direction=direction)
    elif isinstance(curve, (Circle, Ellipse)):
        origin = from_point3d_to_grpc_point(curve.origin)
        reference = from_unit_vector_to_grpc_direction(curve.dir_x)
        axis = from_unit_vector_to_grpc_direction(curve.dir_z)

        if isinstance(curve, Circle):
            grpc_curve = GRPCCurveGeometry(
                origin=origin,
                reference=reference,
                axis=axis,
                radius=GRPCQuantity(
                    value_in_geometry_units=curve.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
                ),
            )
        elif isinstance(curve, Ellipse):
            grpc_curve = GRPCCurveGeometry(
                origin=origin,
                reference=reference,
                axis=axis,
                major_radius=GRPCQuantity(
                    value_in_geometry_units=curve.major_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
                ),
                minor_radius=GRPCQuantity(
                    value_in_geometry_units=curve.minor_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
                ),
            )
    elif isinstance(curve, NURBSCurve):
        grpc_curve = GRPCCurveGeometry(nurbs_curve=from_nurbs_curve_to_grpc_nurbs_curve(curve))
    else:
        raise ValueError(f"Unsupported curve type: {type(curve)}")

    return grpc_curve


def from_nurbs_curve_to_grpc_nurbs_curve(curve: "NURBSCurve") -> GRPCNurbsCurve:
    """Convert a v1 ``NURBSCurve`` to a NURBS curve gRPC message.

    Parameters
    ----------
    curve : NURBSCurve
        Curve to convert.

    Returns
    -------
    GRPCNurbsCurve
        Geometry service gRPC ``NURBSCurve`` message.
    """
    from ansys.api.discovery.v1.design.designmessages_pb2 import (
        ControlPoint as GRPCControlPoint,
        NurbsData as GRPCNurbsData,
    )

    # Convert control points
    control_points = [
        GRPCControlPoint(
            position=from_point3d_to_grpc_point(pt),
            weight=curve.weights[i],
        )
        for i, pt in enumerate(curve.control_points)
    ]

    # Convert nurbs data
    nurbs_data = GRPCNurbsData(
        degree=curve.degree,
        knots=from_knots_to_grpc_knots(curve.knots),
        order=curve.degree + 1,
    )

    return GRPCNurbsCurve(
        control_points=control_points,
        nurbs_data=nurbs_data,
    )


def from_nurbs_surface_to_grpc_nurbs_surface(surface: "NURBSSurface") -> GRPCNurbsSurface:
    """Convert a v1 ``NURBSSurface`` to a NURBS surface gRPC message.

    Parameters
    ----------
    surface : NURBSSurface
        Surface to convert.

    Returns
    -------
    GRPCNurbsSurface
        Geometry service gRPC ``NURBSSurface`` message.
    """
    from ansys.api.discovery.v1.design.designmessages_pb2 import (
        ControlPoint as GRPCControlPoint,
        NurbsData as GRPCNurbsData,
    )

    # Convert control points
    control_points = [
        GRPCControlPoint(
            position=from_point3d_to_grpc_point(point),
            weight=weight,
        )
        for weight, point in zip(surface.weights, surface.control_points)
    ]

    # Convert nurbs data
    nurbs_data_u = GRPCNurbsData(
        degree=surface.degree_u,
        knots=from_knots_to_grpc_knots(surface.knotvector_u),
        order=surface.degree_u + 1,
    )

    nurbs_data_v = GRPCNurbsData(
        degree=surface.degree_v,
        knots=from_knots_to_grpc_knots(surface.knotvector_v),
        order=surface.degree_v + 1,
    )

    return GRPCNurbsSurface(
        control_points=control_points,
        nurbs_data_u=nurbs_data_u,
        nurbs_data_v=nurbs_data_v,
    )


def from_grpc_nurbs_curve_to_nurbs_curve(curve: GRPCNurbsCurve) -> "NURBSCurve":
    """Convert a v1 NURBS curve gRPC message to a ``NURBSCurve``.

    Parameters
    ----------
    curve : GRPCNurbsCurve
        Geometry service gRPC NURBS curve message.

    Returns
    -------
    NURBSCurve
        Resulting converted NURBS curve.
    """
    from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve

    # Extract control points
    control_points = [from_grpc_point_to_point3d(cp.position) for cp in curve.control_points]

    # Extract weights
    weights = [cp.weight for cp in curve.control_points]

    # Extract degree
    degree = curve.nurbs_data.degree

    # Convert gRPC knots to full knot vector
    knots = []
    for grpc_knot in curve.nurbs_data.knots:
        knots.extend([grpc_knot.parameter] * grpc_knot.multiplicity)

    # Create and return the NURBS curve
    return NURBSCurve.from_control_points(
        control_points=control_points,
        degree=degree,
        knots=knots,
        weights=weights,
    )


def from_grpc_nurbs_surface_to_nurbs_surface(surface: GRPCNurbsSurface) -> "NURBSSurface":
    """Convert a v1 NURBS surface gRPC message to a ``NURBSSurface``.

    Parameters
    ----------
    surface : GRPCNurbsSurface
        Geometry service gRPC NURBS surface message.

    Returns
    -------
    NURBSSurface
        Resulting converted NURBS surface.
    """
    from ansys.geometry.core.shapes.surfaces.nurbs import NURBSSurface

    # Extract control points
    control_points = [from_grpc_point_to_point3d(cp.position) for cp in surface.control_points]

    # Extract weights
    weights = [cp.weight for cp in surface.control_points]

    # Extract degree
    degree_u = surface.nurbs_data_u.degree
    degree_v = surface.nurbs_data_v.degree

    # Convert gRPC knots to full knot vector
    knots_u = []
    for grpc_knot in surface.nurbs_data_u.knots:
        knots_u.extend([grpc_knot.parameter] * grpc_knot.multiplicity)
    knots_v = []
    for grpc_knot in surface.nurbs_data_v.knots:
        knots_v.extend([grpc_knot.parameter] * grpc_knot.multiplicity)

    # Create and return the NURBS surface
    return NURBSSurface.from_control_points(
        control_points=control_points,
        degree_u=degree_u,
        degree_v=degree_v,
        knots_u=knots_u,
        knots_v=knots_v,
        weights=weights,
    )


def from_knots_to_grpc_knots(knots: list[float]) -> list[GRPCKnot]:
    """Convert a v1 list of knots to a list of gRPC knot messages.

    Parameters
    ----------
    knots : list[float]
        Source knots data.

    Returns
    -------
    list[GRPCKnot]
        Geometry service gRPC knot messages.
    """
    from collections import Counter

    # Count multiplicities
    multiplicities = Counter(knots)

    # Get unique knots (parameters) in order
    unique_knots = sorted(set(knots))
    knot_multiplicities = [(knot, multiplicities[knot]) for knot in unique_knots]

    # Convert to gRPC knot messages
    grpc_knots = [
        GRPCKnot(
            parameter=knot,
            multiplicity=multiplicity,
        )
        for knot, multiplicity in knot_multiplicities
    ]

    return grpc_knots


def from_grpc_curve_to_curve(curve: GRPCCurveGeometry) -> "Curve":
    """Convert a v1 curve gRPC message to a ``Curve``.

    Parameters
    ----------
    curve : GRPCCurve
        Geometry service gRPC curve message.

    Returns
    -------
    Curve
        Resulting converted curve.
    """
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.shapes.curves.circle import Circle
    from ansys.geometry.core.shapes.curves.ellipse import Ellipse
    from ansys.geometry.core.shapes.curves.line import Line

    origin = from_grpc_point_to_point3d(curve.origin)
    try:
        reference = UnitVector3D([curve.reference.x, curve.reference.y, curve.reference.z])
        axis = UnitVector3D([curve.axis.x, curve.axis.y, curve.axis.z])
    except ValueError:
        # curve will be a line
        pass

    radius = curve.radius.value_in_geometry_units
    major_radius = curve.major_radius.value_in_geometry_units
    minor_radius = curve.minor_radius.value_in_geometry_units
    if radius != 0:
        result = Circle(origin, radius, reference, axis)
    elif major_radius != 0 and minor_radius != 0:
        result = Ellipse(origin, major_radius, minor_radius, reference, axis)
    elif curve.nurbs_curve.nurbs_data.degree != 0:
        result = from_grpc_nurbs_curve_to_nurbs_curve(curve.nurbs_curve)
    elif curve.direction is not None:
        result = Line(
            origin,
            UnitVector3D(
                [
                    curve.direction.x,
                    curve.direction.y,
                    curve.direction.z,
                ]
            ),
        )
    else:
        result = None

    return result


def from_trimmed_surface_to_grpc_trimmed_surface(
    trimmed_surface: "TrimmedSurface",
) -> GRPCTrimmedSurface:
    """Convert a v1 ``TrimmedSurface`` to a trimmed surface gRPC message.

    Parameters
    ----------
    trimmed_surface : TrimmedSurface
        Surface to convert.

    Returns
    -------
    GRPCTrimmedSurface
        Geometry service gRPC ``TrimmedSurface`` message.
    """
    surface_geometry, surface_type = from_surface_to_grpc_surface(trimmed_surface.geometry)

    # In v1, u_min/u_max/v_min/v_max are Quantity messages
    # These are dimensionless parameters
    return GRPCTrimmedSurface(
        surface=surface_geometry,
        type=surface_type,
        u_min=from_parameter_to_grpc_quantity(trimmed_surface.box_uv.interval_u.start),
        u_max=from_parameter_to_grpc_quantity(trimmed_surface.box_uv.interval_u.end),
        v_min=from_parameter_to_grpc_quantity(trimmed_surface.box_uv.interval_v.start),
        v_max=from_parameter_to_grpc_quantity(trimmed_surface.box_uv.interval_v.end),
    )


def from_surface_to_grpc_surface(surface: "Surface") -> tuple[GRPCSurface, GRPCSurfaceType]:
    """Convert a v1 ``Surface`` object to a surface gRPC message.

    Parameters
    ----------
    surface : Surface
        Surface to convert.

    Returns
    -------
    GRPCSurface
        Return ``Surface`` as a ``ansys.api.geometry.Surface`` message.
    GRPCSurfaceType
        Return the grpc surface type of ``Surface``.
    """
    from ansys.geometry.core.shapes.surfaces.cone import Cone
    from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder
    from ansys.geometry.core.shapes.surfaces.plane import PlaneSurface
    from ansys.geometry.core.shapes.surfaces.sphere import Sphere
    from ansys.geometry.core.shapes.surfaces.torus import Torus

    grpc_surface = None
    surface_type = None
    origin = from_point3d_to_grpc_point(surface.origin)
    reference = from_unit_vector_to_grpc_direction(surface.dir_x)
    axis = from_unit_vector_to_grpc_direction(surface.dir_z)

    if isinstance(surface, PlaneSurface):
        grpc_surface = GRPCSurface(origin=origin, reference=reference, axis=axis)
        surface_type = GRPCSurfaceType.SURFACETYPE_PLANE
    elif isinstance(surface, Sphere):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            radius=GRPCQuantity(
                value_in_geometry_units=surface.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            ),
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_SPHERE
    elif isinstance(surface, Cylinder):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            radius=GRPCQuantity(
                value_in_geometry_units=surface.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            ),
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_CYLINDER
    elif isinstance(surface, Cone):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            radius=GRPCQuantity(
                value_in_geometry_units=surface.radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            ),
            half_angle=GRPCQuantity(
                value_in_geometry_units=surface.half_angle.m_as(DEFAULT_UNITS.SERVER_ANGLE)
            ),
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_CONE
    elif isinstance(surface, Torus):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            major_radius=GRPCQuantity(
                value_in_geometry_units=surface.major_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            ),
            minor_radius=GRPCQuantity(
                value_in_geometry_units=surface.minor_radius.m_as(DEFAULT_UNITS.SERVER_LENGTH)
            ),
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_TORUS
    elif isinstance(surface, NURBSSurface):
        grpc_surface = GRPCSurface(
            origin=origin,
            reference=reference,
            axis=axis,
            nurbs_surface=from_nurbs_surface_to_grpc_nurbs_surface(surface),
        )
        surface_type = GRPCSurfaceType.SURFACETYPE_NURBS

    return grpc_surface, surface_type


def from_grpc_surface_to_surface(surface: GRPCSurface, surface_type: "SurfaceType") -> "Surface":
    """Convert a v1 surface gRPC message to a ``Surface`` class.

    Parameters
    ----------
    surface : GRPCSurface
        Geometry service gRPC surface message.

    Returns
    -------
    Surface
        Resulting converted surface.
    """
    from ansys.geometry.core.designer.face import SurfaceType
    from ansys.geometry.core.math.vector import UnitVector3D
    from ansys.geometry.core.shapes.surfaces.cone import Cone
    from ansys.geometry.core.shapes.surfaces.cylinder import Cylinder
    from ansys.geometry.core.shapes.surfaces.plane import PlaneSurface
    from ansys.geometry.core.shapes.surfaces.sphere import Sphere
    from ansys.geometry.core.shapes.surfaces.torus import Torus

    origin = from_grpc_point_to_point3d(surface.origin)
    axis = UnitVector3D([surface.axis.x, surface.axis.y, surface.axis.z])
    reference = UnitVector3D([surface.reference.x, surface.reference.y, surface.reference.z])
    radius = surface.radius.value_in_geometry_units
    major_radius = surface.major_radius.value_in_geometry_units
    minor_radius = surface.minor_radius.value_in_geometry_units

    if surface_type == SurfaceType.SURFACETYPE_CONE:
        result = Cone(origin, radius, surface.half_angle.value_in_geometry_units, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_CYLINDER:
        result = Cylinder(origin, radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_SPHERE:
        result = Sphere(origin, radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_TORUS:
        result = Torus(origin, major_radius, minor_radius, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_PLANE:
        result = PlaneSurface(origin, reference, axis)
    elif surface_type == SurfaceType.SURFACETYPE_NURBS:
        result = from_grpc_nurbs_surface_to_nurbs_surface(surface.nurbs_surface)
    else:
        result = None
    return result


def from_grpc_driving_dimension_to_driving_dimension(
    driving_dimension: GRPCDrivingDimension,
) -> "Parameter":
    """Convert a v1 gRPC driving dimension to a driving dimension object.

    Parameters
    ----------
    driving_dimension : GRPCDrivingDimension
        Source driving dimension type.

    Returns
    -------
    Parameter
        Converted driving dimension.
    """
    from ansys.geometry.core.parameters.parameter import Parameter, ParameterType

    return Parameter(
        id=driving_dimension.id,
        name=driving_dimension.name,
        dimension_type=ParameterType(driving_dimension.dimension_type),
        dimension_value=driving_dimension.dimension_value.value_in_geometry_units,
    )


def from_driving_dimension_to_grpc_driving_dimension(
    driving_dimension: "Parameter",
) -> GRPCDrivingDimension:
    """Convert a v1 driving dimension object to a gRPC driving dimension.

    Parameters
    ----------
    driving_dimension : Parameter
        Source driving dimension type.

    Returns
    -------
    GRPCDrivingDimension
        Converted driving dimension.
    """
    from ansys.geometry.core.parameters.parameter import Parameter

    # Convert Quantity to Real using appropriate units
    value = Parameter.convert_quantity_to_server_units(
        driving_dimension.dimension_value,
        driving_dimension.dimension_type,
    )

    return GRPCDrivingDimension(
        id=driving_dimension.id,
        name=driving_dimension.name,
        dimension_type=driving_dimension.dimension_type.value,
        dimension_value=GRPCQuantity(value_in_geometry_units=value),
    )


def from_grpc_update_status_to_parameter_update_status(
    update_status: GRPCUpdateStatus,
) -> "ParameterUpdateStatus":
    """Convert a v1 gRPC update status to a parameter update status.

    Parameters
    ----------
    update_status : GRPCUpdateStatus
        Source update status.

    Returns
    -------
    ParameterUpdateStatus
        Converted update status.
    """
    from ansys.geometry.core.parameters.parameter import ParameterUpdateStatus

    status_mapping = {
        GRPCUpdateStatus.SUCCESS: ParameterUpdateStatus.SUCCESS,
        GRPCUpdateStatus.FAILURE: ParameterUpdateStatus.FAILURE,
        GRPCUpdateStatus.CONSTRAINED_PARAMETERS: ParameterUpdateStatus.CONSTRAINED_PARAMETERS,
    }
    return status_mapping.get(update_status, ParameterUpdateStatus.UNKNOWN)


def from_design_file_format_to_grpc_file_export_format(
    design_file_format: "DesignFileFormat",
) -> GRPCFileFormat:
    """Convert from a DesignFileFormat object to a gRPC FileExportFormat one.

    Parameters
    ----------
    design_file_format : DesignFileFormat
        The file format desired

    Returns
    -------
    GRPCFileExportFormat
        Converted gRPC File format
    """
    from ansys.geometry.core.designer.design import DesignFileFormat

    if design_file_format == DesignFileFormat.SCDOCX:
        return GRPCFileFormat.FILEFORMAT_SCDOCX
    elif design_file_format == DesignFileFormat.PARASOLID_TEXT:
        return GRPCFileFormat.FILEFORMAT_PARASOLID_TEXT
    elif design_file_format == DesignFileFormat.PARASOLID_BIN:
        return GRPCFileFormat.FILEFORMAT_PARASOLID_BINARY
    elif design_file_format == DesignFileFormat.FMD:
        return GRPCFileFormat.FILEFORMAT_FMD
    elif design_file_format == DesignFileFormat.STEP:
        return GRPCFileFormat.FILEFORMAT_STEP
    elif design_file_format == DesignFileFormat.IGES:
        return GRPCFileFormat.FILEFORMAT_IGES
    elif design_file_format == DesignFileFormat.PMDB:
        return GRPCFileFormat.FILEFORMAT_PMDB
    elif design_file_format == DesignFileFormat.STRIDE:
        return GRPCFileFormat.FILEFORMAT_STRIDE
    elif design_file_format == DesignFileFormat.DISCO:
        return GRPCFileFormat.FILEFORMAT_DISCO
    else:
        return None


def from_material_to_grpc_material(
    material: "Material",
) -> GRPCMaterial:
    """Convert a v1 ``Material`` class to a material gRPC message.

    Parameters
    ----------
    material : Material
        Source material data.

    Returns
    -------
    GRPCMaterial
        Geometry service gRPC material message.
    """
    # Create material properties
    material_properties = []
    for property in material.properties.values():
        prop = GRPCMaterialProperty()
        prop.id.CopyFrom(build_grpc_id(str(property.type.value)))
        prop.display_name = property.name
        prop.value = property.quantity.m
        prop.units = format(property.quantity.units)
        material_properties.append(prop)

    # Create and return the material
    grpc_material = GRPCMaterial()
    grpc_material.name = material.name
    grpc_material.material_properties.extend(material_properties)
    return grpc_material


def from_grpc_matrix_to_matrix(matrix: GRPCMatrix) -> "Matrix44":
    """Convert a v1 gRPC matrix to a matrix.

    Parameters
    ----------
    matrix : GRPCMatrix
        Source gRPC matrix data.

    Returns
    -------
    Matrix44
        Converted matrix.
    """
    import numpy as np

    from ansys.geometry.core.math.matrix import Matrix44

    return Matrix44(
        np.round(
            [
                [matrix.m00, matrix.m01, matrix.m02, matrix.m03],
                [matrix.m10, matrix.m11, matrix.m12, matrix.m13],
                [matrix.m20, matrix.m21, matrix.m22, matrix.m23],
                [matrix.m30, matrix.m31, matrix.m32, matrix.m33],
            ],
            8,
        )
    )


def from_grpc_direction_to_unit_vector(direction: GRPCDirection) -> "UnitVector3D":
    """Convert a v1 gRPC direction to a unit vector.

    Parameters
    ----------
    direction : GRPCDirection
        Source gRPC direction data.

    Returns
    -------
    UnitVector3D
        Converted unit vector.
    """
    from ansys.geometry.core.math.vector import UnitVector3D

    return UnitVector3D([direction.x, direction.y, direction.z])


def from_length_to_grpc_quantity(input: "Distance") -> GRPCQuantity:
    """Convert a v1 ``Distance`` containing a length to a gRPC quantity.

    Parameters
    ----------
    input : Distance
        Source measurement data.

    Returns
    -------
    GRPCQuantity
        Converted gRPC quantity.
    """
    return GRPCQuantity(value_in_geometry_units=input.value.m_as(DEFAULT_UNITS.SERVER_LENGTH))


def from_grpc_quantity_to_distance(input: "GRPCQuantity") -> Distance:
    """Convert a v1 gRPC quantity to ``Distance`` containing a length.

    Parameters
    ----------
    input : GRPCQuantity
        Source measurement data.

    Returns
    -------
    Distance
        Converted Distance quantity.
    """
    return Distance(input.value_in_geometry_units, DEFAULT_UNITS.SERVER_LENGTH)


def from_angle_to_grpc_quantity(input: "Measurement") -> GRPCQuantity:
    """Convert a v1 ``Measurement`` containing an angle to a gRPC quantity.

    Parameters
    ----------
    input : Measurement
        Source measurement data.

    Returns
    -------
    GRPCQuantity
        Converted gRPC quantity.
    """
    return GRPCQuantity(value_in_geometry_units=input.value.m_as(DEFAULT_UNITS.SERVER_ANGLE))


def from_grpc_angle_to_angle(grpc_quantity: GRPCQuantity) -> "pint.Quantity":
    """Convert a v1 gRPC quantity representing an angle to a pint Quantity.

    Parameters
    ----------
    grpc_quantity : GRPCQuantity
        Source gRPC quantity data.

    Returns
    -------
    Measurement
        Converted angle quantity with server angle units.
    """
    return pint.Quantity(grpc_quantity.value_in_geometry_units, DEFAULT_UNITS.SERVER_ANGLE)


def from_area_to_grpc_quantity(input: "Measurement") -> GRPCQuantity:
    """Convert a v1 ``Measurement`` containing an area to a gRPC quantity.

    Parameters
    ----------
    input : Measurement
        Source measurement data.

    Returns
    -------
    GRPCQuantity
        Converted gRPC quantity.
    """
    return GRPCQuantity(value_in_geometry_units=input.value.m_as(DEFAULT_UNITS.SERVER_AREA))


def from_volume_to_grpc_quantity(input: "Measurement") -> GRPCQuantity:
    """Convert a v1 ``Measurement`` containing a volume to a gRPC quantity.

    Parameters
    ----------
    input : Measurement
        Source measurement data.

    Returns
    -------
    GRPCQuantity
        Converted gRPC quantity.
    """
    return GRPCQuantity(value_in_geometry_units=input.value.m_as(DEFAULT_UNITS.SERVER_VOLUME))


def from_grpc_volume_to_volume(grpc_quantity: GRPCQuantity) -> "pint.Quantity":
    """Convert a v1 gRPC quantity representing volume to a pint Quantity.

    Parameters
    ----------
    grpc_quantity : GRPCQuantity
        Source gRPC quantity data.

    Returns
    -------
    pint.Quantity
        Converted volume quantity with server volume units.
    """
    return pint.Quantity(grpc_quantity.value_in_geometry_units, DEFAULT_UNITS.SERVER_VOLUME)


def from_grpc_quantity_to_float(grpc_quantity: GRPCQuantity) -> float:
    """Convert a v1 gRPC quantity to a float.

    Parameters
    ----------
    grpc_quantity : GRPCQuantity
        Source gRPC quantity data.

    Returns
    -------
    float
        The float value contained in the quantity.
    """
    return grpc_quantity.value_in_geometry_units


def from_parameter_to_grpc_quantity(value: float) -> GRPCQuantity:
    """Convert a v1 dimensionless parameter to a gRPC quantity.

    Parameters
    ----------
    value : float
        Dimensionless parameter value (e.g., u/v surface parameters).

    Returns
    -------
    GRPCQuantity
        Converted gRPC quantity with the dimensionless value.
    """
    return GRPCQuantity(value_in_geometry_units=value)


def from_import_options_definitions_to_grpc_import_options_definition(
    import_options_definitions: "ImportOptionsDefinitions",
) -> GRPCImportOptionDefinition:
    """Convert a v1 ``ImportOptionsDefinitions`` to import options definition gRPC message.

    Parameters
    ----------
    import_options_definitions : ImportOptionsDefinitions
        Definition of the import options.

    Returns
    -------
    GRPCImportOptionDefinition
        Geometry service gRPC import options definition message.
    """
    definitions = {}
    for key, definition in import_options_definitions.to_dict().items():
        definitions[key] = GRPCImportOptionDefinition(string_option=str(definition))

    return definitions


def _nurbs_curves_compatibility(backend_version: "semver.Version", grpc_geometries: GRPCGeometries):
    """Check if the backend version is compatible with NURBS curves in sketches.

    Parameters
    ----------
    backend_version : semver.Version
        The version of the backend.
    grpc_geometries : GRPCGeometries
        The gRPC geometries potentially containing NURBS curves.

    Raises
    ------
    GeometryRuntimeError
        If the backend version is lower than 26.1.0 and NURBS curves are present.
    """
    if grpc_geometries.nurbs_curves and backend_version < (26, 1, 0):
        raise GeometryRuntimeError(
            "The usage of NURBS in sketches requires a minimum Ansys release version of "  # noqa: E501
            + "26.1.0, but the current version used is "
            + f"{backend_version}."
        )


def _check_write_body_facets_input(backend_version: "semver.Version", write_body_facets: bool):
    """Check if the backend version is compatible with NURBS curves in sketches.

    Parameters
    ----------
    backend_version : semver.Version
        The version of the backend.
    write_body_facets : bool
        Option to write out body facets.
    """
    if write_body_facets and backend_version < (26, 1, 0):
        from ansys.geometry.core.logger import LOG

        LOG.warning(
            "The usage of write_body_facets requires a minimum Ansys release version of "
            + "26.1.0, but the current version used is "
            + f"{backend_version}."
        )


def from_enclosure_options_to_grpc_enclosure_options(
    enclosure_options: "EnclosureOptions",
) -> GRPCEnclosureOptions:
    """Convert enclosure_options to grpc definition.

    Parameters
    ----------
    enclosure_options : EnclosureOptions
        Definition of the enclosure options.

    Returns
    -------
    GRPCEnclosureOptions
        Grpc converted definition.
    """
    frame = enclosure_options.frame
    return GRPCEnclosureOptions(
        create_shared_topology=enclosure_options.create_shared_topology,
        subtract_bodies=enclosure_options.subtract_bodies,
        frame=from_frame_to_grpc_frame(frame) if frame is not None else None,
        cushion_proportion=enclosure_options.cushion_proportion,
    )


def serialize_body(body: GRPCBodyEntity) -> dict:
    """Serialize a GRPCBodyEntity object into a dictionary.

    It is not directly converted to a pygeometry object because we'll assign it in place and
    construct the object while updating the design object by the tracker output.

    Parameters
    ----------
    body : GRPCBodyEntity
        The gRPC BodyEntity object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the BodyEntity object without gRPC dependencies.
    """
    # Extract basic fields
    body_id = body.id.id
    body_name = body.name
    body_can_suppress = body.can_suppress
    body_master_id = (
        body.master_id.id.id
        if hasattr(body.master_id, "id") and hasattr(body.master_id.id, "id")
        else (body.master_id.id if hasattr(body.master_id, "id") else "")
    )
    body_parent_id = (
        body.parent_id.id.id
        if hasattr(body.parent_id, "id") and hasattr(body.parent_id.id, "id")
        else (body.parent_id.id if hasattr(body.parent_id, "id") else "")
    )
    body_is_surface = body.is_surface

    # Extract transform_to_master matrix
    transform_m00 = body.transform_to_master.m00
    transform_m11 = body.transform_to_master.m11
    transform_m22 = body.transform_to_master.m22
    transform_m33 = body.transform_to_master.m33

    transform_to_master = {
        "m00": transform_m00,
        "m11": transform_m11,
        "m22": transform_m22,
        "m33": transform_m33,
    }

    return {
        "id": body_id,
        "name": body_name,
        "can_suppress": body_can_suppress,
        "transform_to_master": transform_to_master,
        "master_id": body_master_id,
        "parent_id": body_parent_id,
        "is_surface": body_is_surface,
    }


def serialize_component(component: GRPCComponentEntity) -> dict:
    """Serialize a GRPCComponentEntity object into a dictionary.

    Parameters
    ----------
    component : GRPCComponentEntity
        The gRPC ComponentEntity object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the ComponentEntity object without gRPC dependencies.
    """

    def extract_id(obj):
        if hasattr(obj, "id"):
            if hasattr(obj.id, "id"):
                return obj.id.id
            return obj.id
        return ""

    # Extract basic fields
    component_id = component.id.id
    component_name = component.name
    component_display_name = component.display_name

    # Extract part_occurrence
    part_occurrence = None
    if hasattr(component, "part_occurrence"):
        part_occurrence_id = extract_id(component.part_occurrence)
        part_occurrence_name = component.part_occurrence.name
        part_occurrence = {
            "id": part_occurrence_id,
            "name": part_occurrence_name,
        }

    # Extract placement matrix
    placement_m00 = 1.0
    placement_m11 = 1.0
    placement_m22 = 1.0
    placement_m33 = 1.0
    if hasattr(component, "placement"):
        placement_m00 = component.placement.m00
        placement_m11 = component.placement.m11
        placement_m22 = component.placement.m22
        placement_m33 = component.placement.m33

    placement = {
        "m00": placement_m00,
        "m11": placement_m11,
        "m22": placement_m22,
        "m33": placement_m33,
    }

    # Extract part_master
    part_master = None
    if hasattr(component, "part_master"):
        part_master_id = extract_id(component.part_master)
        part_master_name = component.part_master.name
        part_master = {
            "id": part_master_id,
            "name": part_master_name,
        }

    # Extract master_id and parent_id
    master_id = extract_id(component.master_id) if hasattr(component, "master_id") else ""
    parent_id = extract_id(component.parent_id) if hasattr(component, "parent_id") else ""

    return {
        "id": component_id,
        "name": component_name,
        "display_name": component_display_name,
        "part_occurrence": part_occurrence,
        "placement": placement,
        "part_master": part_master,
        "master_id": master_id,
        "parent_id": parent_id,
    }


def serialize_part(part: GRPCPartEntity) -> dict:
    """Serialize a GRPCPartEntity object into a dictionary.

    Parameters
    ----------
    part : GRPCPartEntity
        The gRPC PartEntity object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the PartEntity object without gRPC dependencies.
    """
    return {
        "id": part.id.id,
    }


def serialize_entity_identifier(entity: EntityIdentifier) -> dict:
    """Serialize an EntityIdentifier object into a dictionary.

    Parameters
    ----------
    entity : EntityIdentifier
        The gRPC EntityIdentifier object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the EntityIdentifier object without gRPC dependencies.
    """
    return {
        "id": entity.id,
    }


def serialize_tracked_changes(tracked_changes: GRPCTrackedChanges) -> dict:
    """Serialize a TrackedChanges object into a dictionary.

    Parameters
    ----------
    response : GRPCTrackedChanges
        The gRPC TrackedChanges object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the TrackedChanges object.
    """
    # Extract and serialize parts
    created_parts = [serialize_part(part) for part in getattr(tracked_changes, "created_parts", [])]
    modified_parts = [
        serialize_part(part) for part in getattr(tracked_changes, "modified_parts", [])
    ]
    deleted_parts = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "deleted_parts", [])
    ]

    # Extract and serialize components
    created_components = [
        serialize_component(component)
        for component in getattr(tracked_changes, "created_components", [])
    ]
    modified_components = [
        serialize_component(component)
        for component in getattr(tracked_changes, "modified_components", [])
    ]
    deleted_components = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "deleted_component_ids", [])
    ]

    # Extract and serialize bodies
    created_bodies = [
        serialize_body(body) for body in getattr(tracked_changes, "created_bodies", [])
    ]
    modified_bodies = [
        serialize_body(body) for body in getattr(tracked_changes, "modified_bodies", [])
    ]
    deleted_bodies = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "deleted_body_ids", [])
    ]

    created_faces = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "created_faces", [])
    ]
    modified_faces = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "modified_faces", [])
    ]
    deleted_faces = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "deleted_faces", [])
    ]
    created_edges = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "created_edges", [])
    ]
    modified_edges = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "modified_edges", [])
    ]
    deleted_edges = [
        serialize_entity_identifier(entity)
        for entity in getattr(tracked_changes, "deleted_edges", [])
    ]

    return {
        "created_parts": created_parts,
        "modified_parts": modified_parts,
        "deleted_parts": deleted_parts,
        "created_components": created_components,
        "modified_components": modified_components,
        "deleted_components": deleted_components,
        "created_bodies": created_bodies,
        "modified_bodies": modified_bodies,
        "deleted_bodies": deleted_bodies,
        "created_faces": created_faces,
        "modified_faces": modified_faces,
        "deleted_faces": deleted_faces,
        "created_edges": created_edges,
        "modified_edges": modified_edges,
        "deleted_edges": deleted_edges,
    }


def serialize_tracked_command_response(response: GRPCTrackedCommandResponse) -> dict:
    """Serialize a TrackedCommandResponse object into a dictionary.

    Parameters
    ----------
    response : GRPCTrackedCommandResponse
        The gRPC TrackedCommandResponse object to serialize.

    Returns
    -------
    dict
        A dictionary representation of the TrackedCommandResponse object.
    """
    # Extract command response success status
    success = getattr(response.command_response, "success", False)

    # Extract tracked changes and serialize
    tracked_changes = response.tracked_changes
    serialized = serialize_tracked_changes(tracked_changes)

    serialized["success"] = success
    return serialized


def serialize_repair_command_response(response: GRPCRepairToolResponse) -> dict:
    """Serialize a RepairToolResponse object into a dictionary.

    Parameters
    ----------
    response : GRPCRepairToolResponse
        The gRPC RepairToolResponse object to serialize.
        A dictionary representation of the RepairToolResponse object.
    """
    return {
        "success": response.tracked_command_response.command_response.success,
        "found": getattr(response, "found", -1),
        "repaired": getattr(response, "repaired", -1),
        "tracker_response": serialize_tracked_command_response(response.tracked_command_response),
        "created_bodies_monikers": [
            created_body.id.id
            for created_body in getattr(
                response.tracked_command_response.tracked_changes, "created_bodies", []
            )
        ],
        "modified_bodies_monikers": [
            modified_body.id.id
            for modified_body in getattr(
                response.tracked_command_response.tracked_changes, "modified_bodies", []
            )
        ],
        "deleted_bodies_monikers": [
            deleted_body.id
            for deleted_body in getattr(
                response.tracked_command_response.tracked_changes, "deleted_bodies", []
            )
        ],
    }


def response_problem_area_for_body(response) -> dict:
    """Get a dictionary response from problem areas for a body.

    Parameters
    ----------
    response
        The response to convert.

    Returns
    -------
    dict
        A dictionary representation of the ProblemAreaForBody object.
    """
    return {
        "problems": [
            {"id": res.problem_area_id, "bodies": [body.id for body in res.body_ids]}
            for res in response.result
        ]
    }


def response_problem_area_for_face(response) -> dict:
    """Get a dictionary response from problem areas for a face.

    Parameters
    ----------
    response
        The response to convert.

    Returns
    -------
    dict
        A dictionary representation of the ProblemAreaForFace object.
    """
    return {
        "problems": [
            {"id": res.problem_area_id, "faces": [face.id for face in res.face_ids]}
            for res in response.result
        ]
    }


def response_problem_area_for_edge(response) -> dict:
    """Get a dictionary response from problem areas for an edge.

    Parameters
    ----------
    response
        The response to convert.

    Returns
    -------
    dict
        A dictionary representation of the ProblemAreaForEdge object.
    """
    return {
        "problems": [
            {"id": res.problem_area_id, "edges": [edge.id for edge in res.edge_ids]}
            for res in response.result
        ]
    }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a NURBS surface."""

from functools import cached_property
from typing import TYPE_CHECKING

from beartype import beartype as check_input_types
import numpy as np

from ansys.geometry.core.math import ZERO_POINT3D, Point3D
from ansys.geometry.core.math.constants import UNITVECTOR3D_X, UNITVECTOR3D_Z
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.checks import graphics_required
from ansys.geometry.core.shapes.parameterization import (
    Interval,
    Parameterization,
    ParamForm,
    ParamType,
    ParamUV,
)
from ansys.geometry.core.shapes.surfaces.surface import Surface
from ansys.geometry.core.shapes.surfaces.surface_evaluation import SurfaceEvaluation
from ansys.geometry.core.typing import Real

if TYPE_CHECKING:  # pragma: no cover
    import geomdl.NURBS as geomdl_nurbs  # noqa: N811
    import pyvista as pv


class NURBSSurface(Surface):
    """Represents a NURBS surface.

    Notes
    -----
    This class is a wrapper around the NURBS surface class from the `geomdl` library.
    By leveraging the `geomdl` library, this class provides a high-level interface
    to create and manipulate NURBS surfaces. The `geomdl` library is a powerful
    library for working with NURBS curves and surfaces. For more information, see
    https://pypi.org/project/geomdl/.

    """

    def __init__(
        self,
        origin: Point3D = ZERO_POINT3D,
        reference: UnitVector3D = UNITVECTOR3D_X,
        axis: UnitVector3D = UNITVECTOR3D_Z,
        geomdl_object: "geomdl_nurbs.Surface" = None,
    ):
        """Initialize ``NURBSSurface`` class."""
        try:
            import geomdl.NURBS as geomdl_nurbs  # noqa: N811
        except ImportError as e:  # pragma: no cover
            raise ImportError(
                "The `geomdl` library is required to use the NURBSSurface class. "
                "Please install it using `pip install geomdl`."
            ) from e

        self._nurbs_surface = (
            geomdl_object if geomdl_object else geomdl_nurbs.Surface(normalize_kv=False)
        )
        self._origin = origin
        self._reference = reference
        self._axis = axis

    @property
    def geomdl_nurbs_surface(self) -> "geomdl_nurbs.Surface":
        """Get the underlying `geomdl` NURBS surface object.

        Notes
        -----
        This property gives access to the full functionality of the NURBS surface
        coming from the `geomdl` library. Use with caution.
        """
        return self._nurbs_surface

    @property
    def control_points(self) -> list[Point3D]:
        """Get the control points of the NURBS surface."""
        return [Point3D(pt) for pt in self._nurbs_surface.ctrlpts]

    @property
    def degree_u(self) -> int:
        """Get the degree of the surface in the U direction."""
        return self._nurbs_surface.degree_u

    @property
    def degree_v(self) -> int:
        """Get the degree of the surface in the V direction."""
        return self._nurbs_surface.degree_v

    @property
    def knotvector_u(self) -> list[Real]:
        """Get the knot vector for the u-direction of the surface."""
        return self._nurbs_surface.knotvector_u

    @property
    def knotvector_v(self) -> list[Real]:
        """Get the knot vector for the v-direction of the surface."""
        return self._nurbs_surface.knotvector_v

    @property
    def weights(self) -> list[Real]:
        """Get the weights of the control points."""
        return self._nurbs_surface.weights

    @property
    def origin(self) -> Point3D:
        """Get the origin of the surface."""
        return self._origin

    @property
    def dir_x(self) -> UnitVector3D:
        """Get the reference direction of the surface."""
        return self._reference

    @property
    def dir_z(self) -> UnitVector3D:
        """Get the axis direction of the surface."""
        return self._axis

    @classmethod
    @check_input_types
    def from_control_points(
        cls,
        degree_u: int,
        degree_v: int,
        knots_u: list[Real],
        knots_v: list[Real],
        control_points: list[Point3D],
        weights: list[Real] = None,
        origin: Point3D = ZERO_POINT3D,
        reference: UnitVector3D = UNITVECTOR3D_X,
        axis: UnitVector3D = UNITVECTOR3D_Z,
    ) -> "NURBSSurface":
        """Create a NURBS surface from control points and knot vectors.

        Parameters
        ----------
        degree_u : int
            Degree of the surface in the U direction.
        degree_v : int
            Degree of the surface in the V direction.
        knots_u : list[Real]
            Knot vector for the U direction.
        knots_v : list[Real]
            Knot vector for the V direction.
        control_points : list[Point3D]
            Control points for the surface.
        weights : list[Real], optional
            Weights for the control points. If not provided, all weights are set to 1.
        delta : float, optional
            Evaluation delta for the surface. Default is 0.01.
        origin : Point3D, optional
            Origin of the surface. Default is (0, 0, 0).
        reference : UnitVector3D, optional
            Reference direction of the surface. Default is (1, 0, 0).
        axis : UnitVector3D, optional
            Axis direction of the surface. Default is (0, 0, 1).

        Returns
        -------
        NURBSSurface
            Created NURBS surface.
        """
        nurbs_surface = cls(origin, reference, axis)
        nurbs_surface._nurbs_surface.degree_u = degree_u
        nurbs_surface._nurbs_surface.degree_v = degree_v

        nurbs_surface._nurbs_surface.ctrlpts_size_u = len(knots_u) - degree_u - 1
        nurbs_surface._nurbs_surface.ctrlpts_size_v = len(knots_v) - degree_v - 1

        # If no weights are provided, set all weights to 1.0
        if not weights:
            weights = [1.0] * len(control_points)
        ctrlpts_homogenous = [[*pt, w] for (pt, w) in zip(control_points, weights)]

        nurbs_surface._nurbs_surface.set_ctrlpts(
            ctrlpts_homogenous,
            nurbs_surface._nurbs_surface.ctrlpts_size_u,
            nurbs_surface._nurbs_surface.ctrlpts_size_v,
        )

        nurbs_surface._nurbs_surface.knotvector_u = knots_u
        nurbs_surface._nurbs_surface.knotvector_v = knots_v

        # Verify the surface is valid
        try:
            nurbs_surface._nurbs_surface._check_variables()
        except ValueError as e:
            raise ValueError(f"Invalid NURBS surface: {e}")
        return nurbs_surface

    @classmethod
    @check_input_types
    def fit_surface_from_points(
        cls,
        points: list[Point3D],
        size_u: int,
        size_v: int,
        degree_u: int,
        degree_v: int,
        origin: Point3D = ZERO_POINT3D,
        reference: UnitVector3D = UNITVECTOR3D_X,
        axis: UnitVector3D = UNITVECTOR3D_Z,
    ) -> "NURBSSurface":
        """Fit a NURBS surface to a set of points.

        Parameters
        ----------
        points : list[Point3D]
            Points to fit the surface to.
        size_u : int
            Number of control points in the U direction.
        size_v : int
            Number of control points in the V direction.
        degree_u : int
            Degree of the surface in the U direction.
        degree_v : int
            Degree of the surface in the V direction.
        origin : Point3D, optional
            Origin of the surface. Default is (0, 0, 0).
        reference : UnitVector3D, optional
            Reference direction of the surface. Default is (1, 0, 0).
        axis : UnitVector3D, optional
            Axis direction of the surface. Default is (0, 0, 1).

        Returns
        -------
        NURBSSurface
            Fitted NURBS surface.
        """
        from geomdl import fitting

        converted_pts = [[*pt] for pt in points]

        surface = fitting.interpolate_surface(
            converted_pts,
            size_u=size_u,
            size_v=size_v,
            degree_u=degree_u,
            degree_v=degree_v,
        )

        nurbs_surface = cls(origin, reference, axis)
        nurbs_surface._nurbs_surface.degree_u = degree_u
        nurbs_surface._nurbs_surface.degree_v = degree_v

        nurbs_surface._nurbs_surface.ctrlpts_size_u = surface.ctrlpts_size_u
        nurbs_surface._nurbs_surface.ctrlpts_size_v = surface.ctrlpts_size_v
        nurbs_surface._nurbs_surface.ctrlpts = surface.ctrlpts
        nurbs_surface._nurbs_surface.knotvector = surface.knotvector
        nurbs_surface._nurbs_surface.weights = surface.weights

        return nurbs_surface

    def __eq__(self, other: "NURBSSurface") -> bool:
        """Determine if two surfaces are equal."""
        if not isinstance(other, NURBSSurface):
            return False
        return (
            self._nurbs_surface.degree == other._nurbs_surface.degree
            and self._nurbs_surface.ctrlpts == other._nurbs_surface.ctrlpts
            and self._nurbs_surface.knotvector == other._nurbs_surface.knotvector
            and self._nurbs_surface.weights == other._nurbs_surface.weights
        )

    def parameterization(self) -> tuple[Parameterization, Parameterization]:
        """Get the parametrization of the NURBS surface.

        The parameter is defined in the interval [0, 1] by default. Information
        is provided about the parameter type and form.

        Returns
        -------
        tuple[Parameterization, Parameterization]
            Parameterization in the U and V directions respectively.
        """
        return (
            Parameterization(
                ParamForm.OTHER,
                ParamType.OTHER,
                Interval(
                    start=self._nurbs_surface.domain[0][0], end=self._nurbs_surface.domain[0][1]
                ),
            ),
            Parameterization(
                ParamForm.OTHER,
                ParamType.OTHER,
                Interval(
                    start=self._nurbs_surface.domain[1][0],
                    end=self._nurbs_surface.domain[1][1],
                ),
            ),
        )

    def transformed_copy(self, matrix: Matrix44) -> "NURBSSurface":  # noqa: D102
        raise NotImplementedError("transformed_copy() is not implemented.")

    def evaluate(self, parameter: ParamUV) -> SurfaceEvaluation:
        """Evaluate the surface at the given parameter.

        Parameters
        ----------
        parameter : ParamUV
            Parameter to evaluate the surface at.

        Returns
        -------
        SurfaceEvaluation
            Evaluation of the surface at the given parameter.
        """
        return NURBSSurfaceEvaluation(self, parameter)

    @property
    @graphics_required
    def visualization_polydata(self) -> "pv.PolyData":
        """Get the visualization polydata for the NURBS surface.

        This method generates a triangulated mesh representation of the NURBS surface
        by sampling it at a grid of UV parameters and creating a structured mesh.
        The resolution is automatically determined based on the surface complexity,
        with a minimum of 20 points per direction and additional sampling for
        higher-degree surfaces.

        Returns
        -------
        pv.PolyData
            Triangulated mesh representation of the NURBS surface.

        Notes
        -----
        The surface is sampled in the parametric domain [u_min, u_max] x [v_min, v_max]
        and evaluated at a grid of points. The resulting mesh respects the surface's
        local coordinate system defined by origin, reference (dir_x), and axis (dir_z).
        """
        import pyvista as pv

        # Get the parametric domain from the NURBS surface
        u_domain = self._nurbs_surface.domain[0]  # [u_min, u_max]
        v_domain = self._nurbs_surface.domain[1]  # [v_min, v_max]

        # Determine sampling resolution based on surface complexity
        # Higher degree surfaces need more samples for accurate representation
        base_resolution = 20
        u_samples = max(base_resolution, self.degree_u * 5 + 10)
        v_samples = max(base_resolution, self.degree_v * 5 + 10)

        # Create parametric grid
        u_params = np.linspace(u_domain[0], u_domain[1], u_samples)
        v_params = np.linspace(v_domain[0], v_domain[1], v_samples)

        # Evaluate surface at all UV parameter combinations
        points = []
        for v in v_params:
            for u in u_params:
                # Evaluate the NURBS surface at (u, v)
                point = self._nurbs_surface.evaluate_single((u, v))
                points.append(point)

        # Create structured grid connectivity (quadrilateral faces)
        # Each quad is defined by indices of its 4 corners
        faces = []
        for j in range(v_samples - 1):
            for i in range(u_samples - 1):
                # Calculate vertex indices for the current quad
                # Grid is organized row-by-row (v varies slower than u)
                idx0 = j * u_samples + i
                idx1 = j * u_samples + (i + 1)
                idx2 = (j + 1) * u_samples + (i + 1)
                idx3 = (j + 1) * u_samples + i

                # PyVista face format: [n_points, idx0, idx1, idx2, ...]
                # Create two triangles from each quad for better rendering
                faces.extend([3, idx0, idx1, idx2])  # First triangle
                faces.extend([3, idx0, idx2, idx3])  # Second triangle

        # Create PyVista PolyData from points and faces
        mesh = pv.PolyData(np.array(points), np.array(faces))

        # Apply transformation to align with the surface's local coordinate system
        # Only apply transformation if the surface has non-default orientation/position
        if (
            self._origin != ZERO_POINT3D
            or self._reference != UNITVECTOR3D_X
            or self._axis != UNITVECTOR3D_Z
        ):
            # Build transformation matrix combining rotation and translation
            dir_y = self._axis.cross(self._reference)  # Complete the orthonormal basis

            transform_matrix = np.array(
                [
                    [self._reference.x, dir_y.x, self._axis.x, self._origin.x.m],
                    [self._reference.y, dir_y.y, self._axis.y, self._origin.y.m],
                    [self._reference.z, dir_y.z, self._axis.z, self._origin.z.m],
                    [0, 0, 0, 1],
                ]
            )

            mesh.transform(transform_matrix, inplace=True)

        # Compute normals for better visualization (smooth shading)
        mesh.compute_normals(inplace=True)

        return mesh

    def contains_param(self, param: ParamUV) -> bool:  # noqa: D102
        raise NotImplementedError("contains_param() is not implemented.")

    def contains_point(self, point: Point3D) -> bool:  # noqa: D102
        raise NotImplementedError("contains_point() is not implemented.")

    def project_point(self, point: Point3D) -> SurfaceEvaluation:  # noqa: D102
        raise NotImplementedError("project_point() is not implemented.")


class NURBSSurfaceEvaluation(SurfaceEvaluation):
    """Provides evaluation of a NURBS surface at a given parameter.

    Parameters
    ----------
    nurbs_surface: ~ansys.geometry.core.shapes.surfaces.nurbs.NURBSSurface
        NURBS surface to evaluate.
    parameter: Real
        Parameter to evaluate the NURBS surface at.
    """

    def __init__(self, nurbs_surface: NURBSSurface, parameter: ParamUV) -> None:
        """Initialize the ``NURBSsurfaceEvaluation`` class."""
        self._surface = nurbs_surface
        self._parameter = parameter

        u, v = parameter.u, parameter.v
        domain = nurbs_surface._nurbs_surface.domain
        u_start, u_end = domain[0][0], domain[0][1]
        v_start, v_end = domain[1][0], domain[1][1]

        if not (u_start <= u <= u_end and v_start <= v <= v_end):
            raise ValueError(
                f"Parameter [u={u}, v={v}] is outside the surface domain: "
                f"U[{u_start}, {u_end}], V[{v_start}, {v_end}]"
            )
        self._derivatives = nurbs_surface.geomdl_nurbs_surface.derivatives(u, v, 2)

    @property
    def surface(self) -> "NURBSSurface":
        """Surface being evaluated."""
        return self._surface

    @property
    def parameter(self) -> ParamUV:
        """Parameter the evaluation is based upon."""
        return self._parameter

    @cached_property
    def position(self) -> Point3D:
        """Position of the evaluation.

        Returns
        -------
        Point3D
            Point on the surface at this evaluation.
        """
        return Point3D(self._derivatives[0][0])

    @cached_property
    def normal(self) -> UnitVector3D:
        """Normal to the surface.

        Returns
        -------
        UnitVector3D
            Normal to the surface at this evaluation.
        """
        from geomdl.operations import normal

        uv = [float(self._parameter.u), float(self._parameter.v)]
        result = normal(self._surface.geomdl_nurbs_surface, uv)

        return UnitVector3D(result[1])

    @cached_property
    def u_derivative(self) -> Vector3D:
        """First derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the U parameter at this evaluation.
        """
        return Vector3D(self._derivatives[1][0])

    @cached_property
    def v_derivative(self) -> Vector3D:
        """First derivative with respect to the V parameter.

        Returns
        -------
        Vector3D
            First derivative with respect to the V parameter at this evaluation.
        """
        return Vector3D(self._derivatives[0][1])

    @cached_property
    def uu_derivative(self) -> Vector3D:
        """Second derivative with respect to the U parameter.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U parameter at this evaluation.
        """
        return Vector3D(self._derivatives[2][0])

    @cached_property
    def uv_derivative(self) -> Vector3D:
        """The second derivative with respect to the U and V parameters.

        Returns
        -------
        Vector3D
            Second derivative with respect to the U and V parameters at this evaluation.
        """
        return Vector3D(self._derivatives[1][1])

    @cached_property
    def vv_derivative(self) -> Vector3D:
        """The second derivative with respect to v.

        Returns
        -------
        Vector3D
            Second derivative with respect to the V parameter at this evaluation.
        """
        return Vector3D(self._derivatives[0][2])

    @cached_property
    def min_curvature(self) -> Real:
        """Minimum curvature."""
        raise NotImplementedError("min_curvature() is not implemented.")

    @cached_property
    def min_curvature_direction(self) -> UnitVector3D:
        """Minimum curvature direction."""
        raise NotImplementedError("min_curvature_direction() is not implemented.")

    @cached_property
    def max_curvature(self) -> Real:
        """Maximum curvature."""
        raise NotImplementedError("max_curvature() is not implemented.")

    @cached_property
    def max_curvature_direction(self) -> UnitVector3D:
        """Maximum curvature direction."""
        raise NotImplementedError("max_curvature_direction() is not implemented.")
